import subprocess


class DFTReader:
    @property
    def forces(self):
        """ return numpy.ndarray X x 3 in units [eV/Angstom] """
        raise NotImplementedError()

    @property
    def stress(self):
        """ return numpy.ndarray 3 x 3 in units [GPa] """
        raise NotImplementedError()

    @property
    def energy(self):
        """ return float in units [eV] """
        raise NotImplementedError()

    @property
    def structure(self):
        """ returns pymatgen.Structure. Not units for pymatgen are Angstroms """
        raise NotImplementedError()


class MDReader:
    @property
    def forces(self):
        """ return numpy.ndarray X x 3 in units [eV/Angstom] """
        raise NotImplementedError()

    @property
    def stress(self):
        """ return numpy.ndarray 3 x 3 in units [GPa] """
        raise NotImplementedError()

    @property
    def energy(self):
        """ return float in units [eV] """
        raise NotImplementedError()

    @property
    def structure(self):
        """ returns pymatgen.Structure. Not units for pymatgen are Angstroms """
        raise NotImplementedError()


class MDWriter:
    def write_input(self, directory):
        raise NotImplementedError()


class MDRunner:
    def run(self, writer, command, directory):
        raise NotImplementedError()

    def _run(self, command, run_directory):
        process = subprocess.Popen(command, cwd=run_directory, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        return_code = process.wait()
        stdout, stderr = process.communicate()
        return return_code, stdout, stderr



import pytest
import numpy as np

from dftfit.io.mattoolkit import MTKReader


@pytest.mark.mattoolkit
def test_mattoolkit():
    mtk = MTKReader(2) # super relaxation
    assert np.all(np.isclose(mtk.forces, np.zeros((8, 3))))
    assert np.all(np.isclose(mtk.stress, np.eye(3) * -36.97713))
    assert np.isclose(mtk.energy, -47.47590796)
    structure = mtk.structure
    assert len(structure) == 8
    assert set(s.symbol for s in structure.species) == {'Mg', 'O'}



import sys


class FloatParameter:
    """ Float with tracking. initial value and bounds.

    """
    def __init__(self, initial, bounds=None, fixed=False):
        bounds = bounds or (-sys.float_info.max, sys.float_info.max)
        self.current = float(initial)
        self.bounds = [float(_) for _ in bounds]
        self.computed = None
        self._fixed = fixed

    @property
    def fixed(self):
        return self._fixed or self.computed != None

    def __float__(self):
        if self.computed:
            return self.computed()
        return self.current

    def __str__(self):
        return str(self.current)



from pathlib import Path
from collections import OrderedDict
import subprocess

from lammps.output import LammpsRun, LammpsData
from lammps.inputs import LammpsInput, LammpsScript

from .base import MDReader, MDWriter, MDRunner
from .utils import element_type_to_symbol
from ..potential import Potential



class LammpsReader(MDReader):
    def __init__(self, directory, data_filename='initial.data', log_filename='lammps.log', dump_filename='mol.lammpstrj'):
        self.directory = Path(directory)
        self.data_filename = data_filename
        self.log_filename = log_filename
        self.dump_filename = dump_filename
        self._parse()

    def _parse(self):
        dump_path = list(self.directory.glob('**/%s' % self.dump_filename))
        if len(dump_path) > 1:
            raise ValueError('multiple %s files found within directory' % self.dump_filename)
        elif len(dump_path) == 0:
            raise ValueError('could not find %s file within directory' % self.dump_filename)

        log_path = list(self.directory.glob('**/%s' % self.log_filename))
        if len(log_path) > 1:
            raise ValueError('multiple %s files found within directory' % self.log_filename)
        elif len(log_path) == 0:
            raise ValueError('could not find %s file within directory' % self.log_filename)

        data_path = list(self.directory.glob('**/%s' % self.data_filename))
        if len(data_path) > 1:
            raise ValueError('multiple %s files found within directory' % self.data_filename)
        elif len(data_path) == 0:
            raise ValueError('could not find %s file within directory' % self.data_filename)

        data = LammpsData.from_file(str(data_path[0]))
        output = LammpsRun(str(data_path[0]), lammps_log=str(log_path[0]), lammps_dump=str(dump_path[0]))
        # TODO: check units
        self._forces = output.get_forces(-1)
        self._stress = output.get_stress(-1)
        self._energy = output.get_energy(-1)
        self._structure = output.get_structure(-1)

    @property
    def forces(self):
        return self._forces

    @property
    def stress(self):
        return self._stress

    @property
    def energy(self):
        return self._energy

    @property
    def structure(self):
        return self._structure


lammps_dftfit_set = OrderedDict([
    ('log', 'lammps.log'),
    ('units', 'metal'),
    ('dimension', 3),
    ('boundary', 'p p p'),
    ('atom_style', 'full'),
    ('read_data', 'initial.data'),
    ('kspace_style', []),
    ('pair_style', []),
    ('pair_coeff', []),
    ('set', []),
    ('dump', '1 all custom 1 mol.lammpstrj id type x y z fx fy fz'),
    ('dump_modify', '1 sort id'),
    ('thermo_style', 'custom step etotal pxx pyy pzz pxy pxz pyz'),
    ('run', 0)
])


class LammpsWriter(MDWriter):
    def __init__(self, structure, potential):
        self.structure = structure
        if not isinstance(potential, Potential):
            potential = Potential(potential)

        self.potential = potential.schema # TODO: Hack potential should do more work
        self.data = LammpsData.from_structure(self.structure)
        self.symbol_indicies = {element_type_to_symbol(s): i for s, i in self.data.symbol_indicies.items()}
        lammps_script = LammpsScript(lammps_dftfit_set)
        lammps_script.update([self.charge, self.kspace_style, self.pair_style, self.pair_coeff])
        self.lammps_input = LammpsInput(lammps_script, self.data)

    def write_input(self, directory):
        self.lammps_input.write_input(directory)

    @property
    def charge(self):
        spec = self.potential['spec']
        set_commands = []
        for element, charge in spec.get('charge', {}).items():
            set_commands.append('type %d charge %f' % (self.symbol_indicies[element], float(charge)))
        return ('set', set_commands)

    @property
    def kspace_style(self):
        spec = self.potential['spec']
        if 'kspace' in spec:
            style = spec['kspace']['type']
            tollerance = spec['kspace']['tollerance']
            return ('kspace_style', '%s %f' % (style, float(tollerance)))
        return ('kspace_style', [])

    @property
    def pair_style(self):
        pair_map = {
            'buckingham': 'buck'
        }

        spec = self.potential['spec']
        if 'pair' in spec:
            style = pair_map[spec['pair']['type']]
            cutoff = spec['pair']['cutoff']
            if 'kspace'in spec and spec['kspace']['type'] in {'ewald', 'pppm'}:
                style += '/coul/long'
            return ('pair_style', '%s %f' % (style, float(cutoff)))
        return ('pair_style', [])

    @property
    def pair_coeff(self):
        spec = self.potential['spec']
        if 'pair' in spec:
            symbols_to_indicies = lambda symbols: [self.symbol_indicies[s] for s in symbols]
            pair_coeffs = []
            for coeff in spec['pair']['parameters']:
                pair_coeffs.append(' '.join(list(map(str, symbols_to_indicies(coeff['elements']) + coeff['coefficients']))))
            return ('pair_coeff', pair_coeffs)
        return ('pair_coeff', [])


class LammpsRunner(MDRunner):
    def run(self, writer, command=None, directory=None):
        directory = directory or '.'
        command = command or ['lammps']
        writer.write_input(directory)
        return_code, stdout, stderr = self._run(command, directory)
        if return_code != 0:
            raise ValueError('Lammps calculation exited with non zero return code', stdout, stderr)
        return LammpsReader(directory)



# Taken from marshmallow_polyfield package

from marshmallow import ValidationError
from marshmallow.fields import Field


class PolyField(Field):
    """
    A field that (de)serializes to one of many types. Passed in functions
    are called to disambiguate what schema to use for the (de)serialization
    Intended to assist in working with fields that can contain any subclass
    of a base type
    """
    def __init__(
            self,
            serialization_schema_selector=None,
            deserialization_schema_selector=None,
            many=False,
            **metadata
    ):
        """
        :param serialization_schema_selector: Function that takes in either
        an object representing that object, it's parent object
        and returns the appropriate schema.
        :param deserialization_schema_selector: Function that takes in either
        an a dict representing that object, dict representing it's parent dict
        and returns the appropriate schema

        """
        super(PolyField, self).__init__(**metadata)
        self.many = many
        self.serialization_schema_selector = serialization_schema_selector
        self.deserialization_schema_selector = deserialization_schema_selector

    def _deserialize(self, value, attr, data):
        if not self.many:
            value = [value]

        results = []
        for v in value:
            schema = None
            try:
                schema = self.deserialization_schema_selector(v, data)
                assert hasattr(schema, 'load')
            except Exception:
                schema_message = None
                if schema:
                    schema_message = str(type(schema))

                raise ValidationError(
                    "Unable to use schema. Ensure there is a deserialization_schema_selector"
                    " and that it returns a schema when the function is passed in {value_passed}."
                    " This is the class I got. Make sure it is a schema: {class_type}".format(
                        value_passed=v,
                        class_type=schema_message
                    )
                )
            data, errors = schema.load(v)
            if errors:
                raise ValidationError(errors)
            results.append(data)

        if self.many:
            return results
        else:
            # Will be at least one otherwise value would have been None
            return results[0]

    def _serialize(self, value, key, obj):
        if value is None:
            return None
        try:
            if self.many:
                return [self.serialization_schema_selector(v, obj).dump(v).data for v in value]
            else:
                return self.serialization_schema_selector(value, obj).dump(value).data
        except Exception as err:
            raise TypeError(
                'Failed to serialize object. Error: {0}\n'
                ' Ensure the serialization_schema_selector exists and '
                ' returns a Schema and that schema'
                ' can serialize this value {1}'.format(err, value))



from tempfile import TemporaryDirectory
import os

import numpy as np
from pymatgen import Lattice, Specie, Structure

from dftfit.io import LammpsReader, LammpsWriter
from dftfit.potential import Potential


def test_lammps_reader():
    lammps = LammpsReader('test_files/lammps/mgo')
    assert np.all(np.isclose(lammps.forces, np.zeros((8, 3))))
    assert np.all(np.isclose(lammps.stress, np.eye(3) * 661.01685))
    assert np.isclose(lammps.energy, -85.34101)
    structure = lammps.structure
    assert len(structure) == 8
    assert set(s.symbol for s in structure.species) == {'Mg', 'O'}



def test_lammps_writer():
    supercell = (5, 5, 5)
    a = 4.1990858 # From evaluation of potential
    lattice = Lattice.from_parameters(a, a, a, 90, 90, 90)
    mg = Specie('Mg', 1.4)
    o = Specie('O', -1.4)
    atoms = [mg, o]
    sites = [[0, 0, 0], [0.5, 0.5, 0.5]]
    structure = Structure.from_spacegroup(225, lattice, atoms, sites)

    potential_schema = {
        'version': 'v1',
        'kind': 'Potential',
        'spec': {
            'charge': {
                'Mg': 1.4, 'O': -1.4
            },
            'kspace': {
                'type': 'pppm', 'tollerance': 1e-5
            },
            'pair': {
                'type': 'buckingham',
                'cutoff': 10.0,
                'parameters': [
                    {
                        'elements': ['Mg', 'Mg'],
                        'coefficients': [1309362.2766468062, 0.104, 0.0]
                    },
                    {
                        'elements': ['Mg', 'O'],
                        'coefficients': [9892.357, 0.20199, 0.0]
                    },
                    {
                        'elements': ['O', 'O'],
                        'coefficients': [2145.7345, 0.3, 30.2222]
                    }
                ]
            }
        }
    }

    lammps = LammpsWriter(structure, potential_schema)
    with TemporaryDirectory() as tempdir:
        lammps.write_input(tempdir)
        assert set(os.listdir(tempdir)) == {'lammps.in', 'initial.data'}



import numpy as np
from scipy import optimize

from ._optimize import evaluate, optimize_function


class Dftfit:
    """ DFT Fitting calculations

    Arguments:


    """

    MD_SOLVERS = {'LAMMPS'}

    def __init__(self,
                 weight_forces=0.8, weight_stress=0.1, weight_energy=0.1,
                 max_iterations=100, step_tolerance=1e-6,  method="L-BFGS-B"):
        self.max_iterations = max_iterations
        self.step_tolerance = step_tolerance
        self.method = method

        if not np.isclose(sum([weight_forces, weight_stress, weight_energy]), 1.0, 1e-8):
            raise ValueError('sum of weighting functions must be one')

        self.weights = {
            'forces': weight_forces,
            'stress': weight_stress,
            'energy': weight_energy
        }


    def fit(self, calculations, initial_potential):
        """

        Args:
           - calculations: training set of DFT calculations to fit
           - potential: model of Potential to fit
        """
        def optimization_function(parameters):
            potential = initial_potential.copy()
            potential.optimization_parameters = parameters
            md_calculations = []

            for calculation in calculations:
                md_calculation = evaluate('LAMMPS', calculation.structure, potential)
                md_calculations.append(md_calculation)

            result = optimize_function(md_calculations, calculations, self.weights)
            parameter_str = ' '.join(['{:12.8g}'.format(_) for _ in parameters])
            optimization_str = '%12.6f %12.6f %12.6f %12.6f' % (result['parts']['forces'], result['parts']['stress'], result['parts']['energy'], result['score'])
            print(parameter_str, '|', optimization_str)
            return result['score']

        result = optimize.minimize(
            fun=optimization_function,
            x0=initial_potential.optimization_parameters,
            method=self.method,
            bounds=initial_potential.optimization_bounds,
            tol=self.step_tolerance,
            options={'disp': True, 'maxiter': self.max_iterations}
        )
        print(result)

    def predict(self, structure):
        pass



"""
DFTFIT a package for developing Molecular Dynamics Packages

"""
# stdlib imports
import os
from subprocess import Popen, PIPE
from time import time
from itertools import combinations
from math import sqrt
from copy import deepcopy

# third-party imports
import numpy as np
from scipy.optimize import minimize

# library specific imports
from .io import (
    read_lammps_forces_dump,
    read_lammps_logfile,
    write_lammps_data_file,
    filetypes
)
from shutil import which


class Dftfit:
    """Class responsible for dft optimization

    Configuration variables must be set for optimization with LAMMPS
    to run.

    Default Units
      Force: eV/Angstrom
      Stress: GPa
      Energy: eV
    """

    # Config Definitions
    # lammps_command - command for running lammps execuatable
    # lammps_prefix - arguments for prepending lammps command
    # lammps_postfix - arguments for appending to lammps command
    # lammps_logfile - filename of lammps logfile
    # lammps_datafile - filename for package to write lammps atoms position data file
    # lammps_forces_dumpfile - filename of force dump from lammps
    # lammps_inputfile - filename to of lammps input file
    # lammps_template_files - files that are templated if the format
    #                         dictionary of 'filename': template_string
    #                         each string is formated with str.format() with
    #                         supplied parameters
    # lammps_unit_conversions - to allow for packages to have custom units
    #                     force conversion factor  = (eV/Angstrom)/unit
    #                     stress conversion factor = (GPa)/unit
    #                     energy conversion factor = (eV)/unit
    # lammps_total_energy_scaled - some versions of LAMMPS scale the total energy
    #                              by number of atoms [True/False]
    # weights - how dftfit assigns weights (must add to 1.0). Default only uses forces
    default_config = {
        'lammps_command': 'lmp_ubuntu',
        'lammps_prefix': '',
        'lammps_postfix': '',
        'lammps_logfile': 'log.lammps',
        'lammps_datafile': '',
        'lammps_forces_dumpfile': '',
        'lammps_inputfile': '',
        'lammps_template_files': {},
        'lammps_unit_conversions': {
            'force': 1.0,
            'stress': 1.0,
            'energy': 1.0
        },
        'lammps_total_energy_scaled': False,
        'weights': {
            'forces': 1.0,
            'stresses': 0.0,
            'total-energy': 0.0
        },
        'dftfit_logfile': 'dftfit.log',
        'DEBUG': False
    }

    def __init__(self, config={}):
        self.config = deepcopy(Dftfit.default_config)
        self.config.update(config)

        self.initial_parameters = []
        self.bounds = []
        self.system_configs = []
        self.atom_types = {}
        self.logger = create_logger(
            debug=self.config['DEBUG'],
            log_filename=self.config['dftfit_logfile'])

        self.iteration = {
            'previous_value': None,
            'previous_parameters': None,
            'step': 0,
        }

    def add_system_config(self, filename, filetype):
        """Adds VASP or Quantum Espresso output file
        to system configurations.

        Currently supports:
         - VASP OUTCAR
         - VASP vasprum.xml
         - QE stdout

        It is required that each run output the forces on each atom.
        """
        if filetype not in filetypes.keys():
            raise Exception("Filetype not accepted")

        system_config = filetypes[filetype](filename)
        system_config.update({'filename': filename})

        self.system_configs.append(system_config)

    def _validate_run(self):
        """Validates that dftfit run is setup to run correctly

        """
        self.logger.info("validating configuration for run")

        # Check that there are system configurations
        if len(self.system_configs) == 0:
            error_str = 'Error: [config] must add system configurations "add_system_config"'
            raise Exception(error_str)

        # run validation
        # Check that a lammps input file exists
        if (self.config['lammps_inputfile'] not in self.config['lammps_template_files'].keys() and
            not os.path.isfile(self.config['lammps_inputfile'])):
            error_str = (
                '[config] lammps input file not found and not in template files "{0}"'
            ).format(self.config['lammps_inputfile'])
            self.logger.error(error_str)
            raise Exception(error_str)

        # Check that the number of initial parameters matches bounds
        if len(self.initial_parameters) != len(self.bounds):
            error_str = 'number of parameters must match bounds'
            self.logger.error(error_str)
            raise Exception(error_str)

        # Check that parameters are within the bounds
        for i, (parameter, (lower_bound, upper_bound)) in enumerate(zip(
                self.initial_parameters, self.bounds)):
            if not lower_bound < parameter < upper_bound:
                error_str = (
                    '{} parameter not within bounds.'
                    '{:.3g} < {:.3g} < {:.3g} not satisfied'
                ).format(i, lower_bound, parameter, upper_bound)
                self.logger.error(error_str)
                raise Exception(error_str)

        # Warn user if no template files are setup
        if len(self.config['lammps_template_files']) == 0:
            self.logger.warning("no template files supplied for dftfit")

        # Warn User if weights are zero
        for unit, weight in self.config['weights'].items():
            if weight == 0.0:
                warning_str = (
                    "{} will not be used in optimization since weight is zero"
                ).format(unit)
                self.logger.warning(warning_str)

        # If total-energy is non-zero at least 2 configurations are required
        if self.config['weights']['total-energy'] != 0.0 and \
           len(self.system_configs) < 2:
            error_str = (
                "at least 2 system configurations required if "
                "total-energy weight is non-zero"
            )
            self.logger.error(error_str)
            raise Exception(error_str)

    def _validate_config(self):
        """Validates that dftfit configuration is correct

        """
        self.logger.info("validating configuration from initialization")

        # Check if lammps execuate is in path or absolute path is given
        if which(self.config['lammps_command']) is None:
            error_str = (
                '[config] lammps execuatable not found "{0}"'
            ).format(self.config['lammps_command'])
            self.logger.error(error_str)
            raise Exception(error_str)

        # Check that user has supplied an datafilename
        if self.config['lammps_datafile'] is '':
            error_str = '[config] user must supply a name for lammps datafile'
            self.logger.error(error_str)
            raise Exception(error_str)

        # Check that user has supplied an dumpfile
        if self.config['lammps_forces_dumpfile'] is '':
            error_str = '[config] user must supply a name for lammps dumpfile'
            self.logger.error(error_str)
            raise Exception(error_str)

        # Check that unit conversions are non-zero
        for unit, conversion in self.config['lammps_unit_conversions'].items():
            if conversion == 0.0:
                error_str = (
                    "unit {} conversion factor must be non-zero"
                ).format(unit)
                self.logger.error(error_str)
                raise Exception(error_str)

    def _run_lammps(self):
        """Runs LAMMPS executable on given input file.

        Notice that this executable is *not* run in the shell. Thus
        environment varaibles will not be available. See
        Popen(shell=False) doc. The LAMMPS '-i' flag is not used
        becuase older versions of LAMMPS do not support this argument.

        """
        with open(self.config['lammps_inputfile']) as f:
            lammps_input = f.read()

        # TODO split works as long as there are no spaces in each
        # argument
        lammps_command = self.config['lammps_command'].split()
        lammps_prefix = self.config.get('lammps_prefix', "").split()
        lammps_postfix = self.config.get('lammps_postfix', "").split()

        proc = Popen(lammps_prefix + lammps_command + lammps_postfix,
                     stdin=PIPE, stdout=PIPE, stderr=PIPE)

        lammps_output = proc.communicate(lammps_input.encode())
        proc.wait()

        stdout = lammps_output[0].decode()
        stderr = lammps_output[1].decode()

        if proc.returncode != 0:
            error_str = (
                "Lammps run exited {} check lammps logfile:\n{}"
            ).format(proc.returncode, stderr)
            self.logger.error(error_str)
            raise Exception(error_str)

    def _calculate_lammps(self, structure, atom_types, parameters):
        """Glue function to write input filesm, run lammps, and collect the
        forces, stress, and total_energy from a given static
        calculation.

        Collects (if resepective weight is non-zero):
         - Forces
         - Stresses
         - Total Energy

        The LAMMPS data file has all of the atomic coordinates along
        with lattice dimmensions. DFTFIT considers all other files
        required by lammps as templates. Templates must currently be
        supplied as strings.

        """
        # Write the atomic positions file
        write_lammps_data_file(structure, atom_types, self.config['lammps_datafile'])

        # Write Lammps Template Files (fill with parameters)
        for filename, template in self.config['lammps_template_files'].items():
            with open(filename, "w") as f:
                f.write(template.format(*parameters))

        self._run_lammps()

        # TODO should we really read the last value
        thermo = read_lammps_logfile(self.config['lammps_logfile'])['thermo']
        stresses = np.array(
            [[thermo['Pxx'][-1], thermo['Pxy'][-1], thermo['Pxz'][-1]],
             [thermo['Pxy'][-1], thermo['Pyy'][-1], thermo['Pyz'][-1]],
             [thermo['Pxz'][-1], thermo['Pyz'][-1], thermo['Pzz'][-1]]])

        total_energy = thermo['TotEng'][-1]
        if self.config['lammps_total_energy_scaled']:
            total_energy *= structure.get_number_of_atoms()

        forces = read_lammps_forces_dump(self.config['lammps_forces_dumpfile'])

        return {
            'forces': forces * self.config['lammps_unit_conversions']['force'],
            'stresses': stresses * self.config['lammps_unit_conversions']['stress'],
            'total-energy': total_energy * self.config['lammps_unit_conversions']['energy'],
        }

    def __str__(self):
        config_str = (
            "\nLammps Configuration:\n"
            "  command   : {0}\n"
            "  inputfile : {1}\n"
            "  logfile   : {2}\n"
            "  datafile  : {3}\n"
            "  force dump: {4}\n"
        ).format(
            self.config['lammps_prefix'] + ' ' + self.config['lammps_command'] + ' ' + self.config['lammps_postfix'],
            self.config['lammps_inputfile'],
            self.config['lammps_logfile'],
            self.config['lammps_datafile'],
            self.config['lammps_forces_dumpfile'])

        template_str = "Templated Files\n"
        for filename, template in self.config['lammps_template_files'].items():
            template_str += "  {0}\n".format(filename)

        parameters_str = "Parameters:\n"
        for i, (parameter, bound) in enumerate(zip(self.initial_parameters, self.bounds), start=1):
            parameters_str += (
                "Parameter {0:>2}: initial value: {1}\n"
                "              bounds: {2:.4g} <-> {3:.4g}\n"
            ).format(i, parameter, *bound)

        system_configs_str = (
            "Number of system configurations: {0}\n"
            "filename\tchemical formula\tnumber of atoms\n"
        ).format(len(self.system_configs))
        for system_config in self.system_configs:
            system_configs_str += "{0}\t{1}\t{2}\n".format(
                system_config.get('filename'),
                system_config['system'].get_chemical_formula(),
                len(system_config['system']))

        return "\n".join([config_str,
                          template_str,
                          parameters_str,
                          system_configs_str])

    def __repr__(self):
        return "<DFTFIT: {0} system configs {1} parameters>".format(
            len(self.system_configs), len(self.initial_parameters))

    def optimize(self, package='scipy'):
        """
        High level routine to optimize dft structure
        """
        self._validate_run()
        self.logger.info(self)

        # Calculate Least Square Error according to
        # http://dx.doi.org/10.1063/1.1513312
        # forces / stresses / total-energy
        def optimize_function(p, grad=None):
            # n_ -> numerator term
            # d_ -> denomenator term
            n_force_sq_error = 0.0
            d_force_sq_error = 0.0
            n_stress_sq_error = 0.0
            d_stress_sq_error = 0.0
            n_energy_sq_error = 0.0
            d_energy_sq_error = 0.0

            # Run MD for each system configuration
            for system_config in self.system_configs:
                if self.config.get('DEBUG'):
                    start_time = time()

                system_config.update({
                    'md': self._calculate_lammps(
                        system_config['system'], self.atom_types, p)
                })

                if self.config.get('DEBUG'):
                    self.logger.debug("LAMMPS Calculation time: {0}".format(time() - start_time))

                md_forces = system_config['md']['forces']
                dft_forces = system_config['dft']['forces']
                n_force_sq_error += np.sum((md_forces - dft_forces)**2.0)
                d_force_sq_error += np.sum(dft_forces**2.0)

                self.logger.debug((
                    "\nDFT force:\n {}\n"
                    "MD force:\n {}\n"
                ).format(dft_forces, md_forces))

                md_stresses = system_config['md']['stresses']
                dft_stresses = system_config['dft']['stresses']
                n_stress_sq_error += np.sum((md_stresses - dft_stresses)**2.0)
                d_stress_sq_error += np.sum(dft_stresses**2.0)

                self.logger.debug((
                    "\nDFT stresses:\n {}\n"
                    "MD stresses:\n {}\n"
                ).format(dft_stresses, md_stresses))

            for system_config_i, system_config_j in combinations(self.system_configs, 2):
                md_energy_i = system_config_i['md']['total-energy']
                md_energy_j = system_config_j['md']['total-energy']

                dft_energy_i = system_config_i['dft']['total-energy']
                dft_energy_j = system_config_j['dft']['total-energy']

                n_energy_sq_error += ((md_energy_i - md_energy_j) - (dft_energy_i - dft_energy_j))**2.0
                d_energy_sq_error += (dft_energy_i - dft_energy_j)**2.0

            w_force = self.config['weights']['forces']
            w_stress = self.config['weights']['stresses']
            w_energy = self.config['weights']['total-energy']

            force_sq_error = sqrt(n_force_sq_error / d_force_sq_error)
            stress_sq_error = sqrt(n_stress_sq_error / d_stress_sq_error)

            # TODO hack to allow one system config if energy weight is zero
            if len(self.system_configs) == 1 and \
               self.config['weights']['total-energy'] == 0.0:
                energy_sq_error = 0.0
            else:
                energy_sq_error = sqrt(n_energy_sq_error / d_energy_sq_error)

            sq_error = (w_force * force_sq_error +
                        w_stress * stress_sq_error +
                        w_energy * energy_sq_error)

            self.logger.debug((
                "\nForce  SQ Error : {} Weight: {}\n"
                "Stress SQ Error : {} Weight: {}\n"
                "Energy SQ Error : {} Weight: {}\n"
                "Total  SQ Error : {}"
            ).format(force_sq_error, w_force,
                     stress_sq_error, w_stress,
                     energy_sq_error, w_energy,
                     sq_error))

            self.logger.info((
                'Step: {:3d} Value: {:>20.16} Parameters: {}'
            ).format(self.iteration['step'], sq_error, p))

            self.iteration['previous_parameters'] = p.copy()
            self.iteration['previous_value'] = sq_error
            self.iteration['step'] += 1

            return sq_error

        # Choose a python non-linear optimization solver (scipy)
        # See http://docs.scipy.org/doc/scipy/reference/optimize.html
        # For detailed informaion
        # def optimize_message(p):
        #     print("Current Iteration Parameters: {0}".format(p))
        #
        if package == 'scipy':
            minimize(
                fun=optimize_function,
                x0=self.initial_parameters,
                method="L-BFGS-B",
                #method="powell",
                #method="COBYLA",
                #method="TNC",
                #method="SLSQP",
                jac=False,
                bounds=self.bounds,
                options={'disp': True},
            )
        elif package == 'nlopt':
            # NLOPT optimization solver
            # used because reference exists that shows algorithm works
            # Using BOBYQA algorithm
            # http://ab-initio.mit.edu/wiki/index.php/NLopt_Algorithms#BOBYQA
            opt = nlopt.opt(nlopt.LN_BOBYQA, len(self.initial_parameters))
            lower_bounds, upper_bounds = zip(*self.bounds)

            self.logger.info((
                "Using {} optimization algorithm with {} parameters\n"
                "{}\n"
            ).format(opt.get_algorithm(), opt.get_dimension(), opt.get_algorithm_name()))

            opt.set_lower_bounds(lower_bounds)
            opt.set_upper_bounds(upper_bounds)

            self.logger.info((
                "\nBounded Problem:\n"
                "Lower Bound: {}\n"
                "Upper Bound: {}\n"
            ).format(opt.get_lower_bounds(), opt.get_upper_bounds()))

            opt.set_min_objective(optimize_function)
            opt.set_ftol_rel(1e-6)
            # opt.set_maxeval(0) # 0 or negative for no limit
            # opt.set_maxtime(0) # seconds (0 or negative for no limit)
            optimized_parameters = opt.optimize(list(self.initial_parameters))
            min_objective_value = opt.last_optimum_value()

            self.logger.info((
                "\nOptimized Parameters: {}\n"
                "Minimum Value: {}\n"
            ).format(optimized_parameters, min_objective_value))
        else:
            raise ValueError('Optimzation Packages %s not recognized' % package)



""" Handles collection of trial and testing data.

Has a caching layer as to speed up future runs
"""
import json

import yaml

from .schema import TrainingSchema
from .io import MTKReader


class Training:
    def __init__(self, schema):
        schema_load, errors = TrainingSchema().load(schema)
        self.schema = schema_load
        self._gather_calculations()

    def _gather_calculations(self):
        self._calculations = []
        for calculation in self.schema['spec']:
            if calculation['type'] == 'mattoolkit':
                self._calculations.extend(self.download_mattoolkit_calculations(calculation['selector']))

    @property
    def calculations(self):
        return self._calculations

    def __iter__(self):
        return iter(self._calculations)

    def download_mattoolkit_calculations(self, selector):
        from mattoolkit.api import CalculationResourceList

        calculations = CalculationResourceList()
        calculations.get(params={'labels': selector['labels']})
        return [MTKReader(c.id) for c in calculations.items]

    @classmethod
    def from_file(cls, filename, format=None):
        if format not in {'json', 'yaml'}:
            if filename.endswith('json'):
                format = 'json'
            elif filename.endswith('yaml') or filename.endswith('yml'):
                format = 'yaml'
            else:
                raise ValueError('unrecognized filetype from filename %s' % filename)

        if format == 'json':
            with open(filename) as f:
                return cls(json.load(f))
        elif format in {'yaml', 'yml'}:
            with open(filename) as f:
                return cls(yaml.load(f))

    def __str__(self):
        return json.dumps(self.schema, sort_keys=True, indent=4)



import click

from . import cli

from ..dftfit import Dftfit
from ..potential import Potential
from ..training import Training


@cli.command()
@click.option('-t', '--training', type=click.Path(exists=True))
@click.option('-p', '--potential', type=click.Path(exists=True))
@click.pass_context
def train(ctx, training, potential):
    dftfit = Dftfit()
    dftfit.fit(Training.from_file(training), Potential.from_file(potential))



from .base import DFTReader

import numpy as np


class MTKReader(DFTReader):
    def __init__(self, calculation_id):
        self.calculation_id = calculation_id
        self._load()

    def _load(self):
        from mattoolkit.api.calculation import CalculationResourceItem
        calculation = CalculationResourceItem(self.calculation_id)
        calculation.get()

        if calculation.format in {'VASP'}:
            results = calculation.results
            self._forces = np.array(results.final_forces)
            self._stress = np.array(results.final_stress) * 1e3 # kbar -> bar
            self._energy = float(results.final_energy)
            self._structure = results.final_structure
        else:
            raise ValueError('unable to use calculation type %s' % calculation.type)

    @property
    def forces(self):
        return self._forces

    @property
    def stress(self):
        return self._stress

    @property
    def energy(self):
        return self._energy

    @property
    def structure(self):
        return self._structure



import click
import logging

from ..logging import LOG_LEVELS, init_logging


@click.group()
@click.option('--loglevel', type=click.Choice(LOG_LEVELS), default='WARNING')
@click.pass_context
def cli(ctx, loglevel):
    init_logging(loglevel)


from . import train



from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand
# To use a consistent encoding
from codecs import open
from os import path
import sys
import shlex


class PyTest(TestCommand):
    user_options = [('pytest-args=', 'a', "Arguments to pass to py.test")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = []

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        #import here, cause outside the eggs aren't loaded
        import pytest
        if isinstance(self.pytest_args, str):
            self.pytest_args = shlex.split(self.pytest_args)
        errno = pytest.main(self.pytest_args)
        sys.exit(errno)

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()


version='0.0.4'
setup(
    name='dftfit',
    version=version,
    description='Ab-Initio Molecular Dynamics Potential Development',
    long_description=long_description,
    url='https://gitlab.aves.io/costrouc/dftfit',
    author='Chris Ostrouchov',
    author_email='chris.ostrouchov+dftfit@gmail.com',
    classifiers=[
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    cmdclass = {'test': PyTest},
    keywords='materials dft molecular dynamics lammps science hpc',
    download_url='https://gitlab.aves.io/costrouc/dftfit/repository/archive.zip?ref=v%s' % version,
    packages=find_packages(exclude=('tests', 'docs', 'notebooks', 'examples')),
    install_requires=['pymatgen==2017.7.4', 'pymatgen-lammps', 'marshmallow', 'pyyaml', 'click'],
    extras_require={
        'mattoolkit': 'mattoolkit'
    },
    tests_require=['pytest'],
    entry_points={
        'console_scripts': [
            'dftfit=dftfit.__main__:main'
        ]
    },
    license='LGPLv2.1+',
)



""" A move to make Potentials as universal as possible


"""
import json

import yaml
import numpy as np
from pymatgen.core import Composition

from .schema import PotentialSchema
from .parameter import FloatParameter


class Potential:
    def __init__(self, schema):
        schema_load, errors = PotentialSchema().load(schema)
        self.schema = schema_load
        self._apply_constraints()
        self._collect_parameters()

    def _apply_constraints(self):
        for constraint, value in self.schema['spec'].get('constraint', {}).items():
            if constraint == 'charge_balance':
                composition = Composition(value)
                charges = self.schema['spec'].get('charge', {})
                if not {e.symbol for e in composition.keys()} <= charges.keys():
                    raise ValueError('charge ballance constrains requires all elements to be defined in charge')
                for charge_element, parameter in charges.items():
                    if isinstance(parameter, FloatParameter) and not parameter.fixed:
                        break
                else:
                    if abs(sum(float(charges[element.symbol]) * amount for element, amount in composition.items())) > 1e-8:
                        raise ValueError('no parameters to apply charge constraint and charge does ballance')
                    continue
                parameter.computed = lambda: -sum(float(charges[element.symbol]) * amount for element, amount in composition.items() if element.symbol != charge_element)

    def _collect_parameters(self):
        self._parameters = []

        def _walk(value): # Ordered traversal of dictionary
            if isinstance(value, dict):
                for key in sorted(value.keys()):
                    _walk(value[key])
            elif isinstance(value, (tuple, list)):
                for item in value:
                    _walk(item)
            elif isinstance(value, FloatParameter):
                self._parameters.append(value)
        _walk(self.schema)

        self._optimization_parameters = [p for p in self._parameters if not p.fixed]

    @classmethod
    def from_file(cls, filename, format=None):
        if format not in {'json', 'yaml'}:
            if filename.endswith('json'):
                format = 'json'
            elif filename.endswith('yaml') or filename.endswith('yml'):
                format = 'yaml'
            else:
                raise ValueError('unrecognized filetype from filename %s' % filename)

        if format == 'json':
            with open(filename) as f:
                return cls(json.load(f))
        elif format in {'yaml', 'yml'}:
            with open(filename) as f:
                return cls(yaml.load(f))

    def as_dict(self):
        schema_dump, errors = PotentialSchema().dump(self.schema)
        return schema_dump

    def __copy__(self):
        return type(self)(self.as_dict())

    def copy(self):
        return self.__copy__()

    def __hash__(self):
        class CustomEncoder(json.JSONEncoder):
            def default(self, obj):
                if isinstance(obj, FloatParameter):
                    return "FloatParameter"
                else:
                    return super().default(obj)
        return hash(json.dumps(self.schema, sort_keys=True, cls=CustomEncoder))

    def __eq__(self, other):
        return hash(self) == hash(other) and np.all(np.isclose(self.parameters, other.parameters, rtol=1e-16))

    @property
    def parameters(self):
        """ Returns parameters for potentials as a list of float values

        """
        return np.array([float(parameter) for parameter in self._parameters])

    @property
    def optimization_parameters(self):
        return np.array([float(parameter) for parameter in self._optimization_parameters])

    @optimization_parameters.setter
    def optimization_parameters(self, parameters):
        """ Update potential with given parameters

        """
        if len(parameters) != len(self._optimization_parameters):
            raise ValueError('updating parameters does not match length of potential parameters')

        for parameter, update_parameter in zip(self._optimization_parameters, parameters):
            parameter.current = float(update_parameter)

    @property
    def optimization_bounds(self):
        return np.array([parameter.bounds for parameter in self._optimization_parameters])

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return json.dumps(self.as_dict(), sort_keys=True, indent=4)



import numpy as np

from dftfit.io import VaspReader


def test_vasp_reader():
    calculation = VaspReader('test_files/vasp', vasprun_filename='vasprun.xml.mgo')
    assert np.all(np.isclose(calculation.energy, -11.87675826))
    assert np.all(np.isclose(calculation.stress, np.eye(3) * -12.62984))
    assert np.all(np.isclose(calculation.forces, np.zeros((2, 3))))
    structure = calculation.structure
    assert len(structure) == 2
    assert set(s.symbol for s in structure.species) == {'Mg', 'O'}



from itertools import combinations
from tempfile import TemporaryDirectory
import math

import numpy as np

from .io.lammps import LammpsWriter, LammpsRunner


def evaluate(runner, structure, potential):
    if runner == 'LAMMPS':
        with TemporaryDirectory() as tempdir:
            writer = LammpsWriter(structure, potential)
            return LammpsRunner().run(writer, ['lammps', '-i', 'lammps.in'], tempdir)
    else:
        raise ValueError('unknown runner')


def optimize_function(md_calculations, dft_calculations, weights):
    n_force_sq_error = 0.0
    d_force_sq_error = 0.0
    n_stress_sq_error = 0.0
    d_stress_sq_error = 0.0
    n_energy_sq_error = 0.0
    d_energy_sq_error = 0.0

    for md_calculation, dft_calculation in zip(md_calculations, dft_calculations):
        n_force_sq_error += np.sum((md_calculation.forces - dft_calculation.forces)**2.0)
        d_force_sq_error += np.sum(dft_calculation.forces**2.0)

        n_stress_sq_error += np.sum((md_calculation.stress - dft_calculation.stress)**2.0)
        d_stress_sq_error += np.sum(dft_calculation.stress**2.0)

    for (md_calc_i, dft_calc_i), (md_calc_j, dft_calc_j) in combinations(zip(md_calculations, dft_calculations), 2):
        n_energy_sq_error += ((md_calc_i.energy - md_calc_j.energy) - (dft_calc_i.energy - dft_calc_j.energy))**2.0
        d_energy_sq_error += (dft_calc_i.energy - dft_calc_j.energy)**2.0

    force_sq_error = math.sqrt(n_force_sq_error / d_force_sq_error)
    stress_sq_error = math.sqrt(n_stress_sq_error / d_stress_sq_error)
    energy_sq_error = math.sqrt(n_energy_sq_error / d_energy_sq_error)

    score = (
        weights['forces'] * force_sq_error + \
        weights['stress'] * stress_sq_error + \
        weights['energy'] * energy_sq_error
    )

    return {
        'weights': {'forces': weights['forces'], 'stress': weights['stress'], 'energy': weights['energy']},
        'parts': {'forces': force_sq_error, 'stress': stress_sq_error, 'energy': energy_sq_error},
        'score': score
    }


# def optimize():
#     # where to optimize potential based on condition
#     if package == 'scipy':
#         minimize(
#             fun=optimize_function,
#             x0=self.initial_parameters,
#             method="L-BFGS-B",
#             #method="powell",
#             #method="COBYLA",
#             #method="TNC",
#             #method="SLSQP",
#             jac=False,
#             bounds=self.bounds,
#             options={'disp': True},
#         )
#     elif package == 'nlopt':
#         # NLOPT optimization solver
#         # used because reference exists that shows algorithm works
#         # Using BOBYQA algorithm
#         # http://ab-initio.mit.edu/wiki/index.php/NLopt_Algorithms#BOBYQA
#         opt = nlopt.opt(nlopt.LN_BOBYQA, len(self.initial_parameters))
#         lower_bounds, upper_bounds = zip(*self.bounds)

#         self.logger.info((
#             "Using {} optimization algorithm with {} parameters\n"
#             "{}\n"
#         ).format(opt.get_algorithm(), opt.get_dimension(), opt.get_algorithm_name()))

#         opt.set_lower_bounds(lower_bounds)
#         opt.set_upper_bounds(upper_bounds)

#         self.logger.info((
#             "\nBounded Problem:\n"
#             "Lower Bound: {}\n"
#             "Upper Bound: {}\n"
#         ).format(opt.get_lower_bounds(), opt.get_upper_bounds()))

#         opt.set_min_objective(optimize_function)
#         opt.set_ftol_rel(1e-6)
#         # opt.set_maxeval(0) # 0 or negative for no limit
#         # opt.set_maxtime(0) # seconds (0 or negative for no limit)
#         optimized_parameters = opt.optimize(list(self.initial_parameters))
#         min_objective_value = opt.last_optimum_value()
#     else:
#         raise ValueError('Optimzation Packages %s not recognized' % package)
#     pass
